from KINCluster.core.item import Item
from KINCluster.lib.tokenizer import stemize, tagging, is_noun
from KINCluster.lib.stopwords import stopwords
from KINCluster import settings

from typing import List, Union, Any
from math import log10 as normalize
from collections import Counter, OrderedDict
from itertools import chain
from functools import reduce
import re

# type hinting
itemID = int # item id
dID = int # document id
wID = str # word id, it str cuz word counter, {word: count}

import numpy as np

class extractable:
    """extractable is decorater for extractor, which extract something.

extractor.dump() return all result of extractable.
extractable function must have argument(iid: itemID)
    """
    s = {}
    def __init__(self, func):
        extractable.s[func.__name__] = func
        self.func = func

    def __get__(self, obj, klass=None):
        def _call_(*args, **kwargs):
            return self.func(obj, *args, **kwargs)
        return _call_

    @staticmethod
    def help():
        """print extractable function documents
        """
        print ('extractable', 'help')
        for f in extractable.s.values():
            print (f.__name__, f.__doc__)

class Extractor:
    """Extractor is extract feature from cluster

Usage:
    Extractor(cluster[, tokenizer])
    
    dump extract all of extractable feature as dict.

Default extractable:
    items       : items of clustered dump
    vectors     : vectors of clustered dump
    counter     : counter(split by tokenizer) of clustered dump
    center      : return center of vectors, (using np.mean)
    """
    def __init__(self, cluster, tokenizer=stemize, notword='[^a-zA-Z가-힣0-9]'):
        self.tokenizer = tokenizer
        self.not_word = re.compile(notword)
        self.__c = cluster

    @staticmethod
    def __find_center(vectors: np.ndarray) -> int:
        if isinstance(vectors, list):
            h = len(vectors)
            w = len(list(chain.from_iterable(vectors))) / h
        else:
            h, w = vectors.shape
        mean = np.mean(vectors)
        return mean, int(np.abs(vectors - mean).argmin() / w)

    def __get_word_count(self, items: List[Item]) -> Union[set, List[Counter]]:
        counter = Counter()
        for idx, item in enumerate(items):
            for word in self.tokenizer(repr(item)):
                if not word in stopwords() and not self.not_word.search(word):
                    counter[word] += 1
        return counter

    def dump(self, iid: itemID) -> Item:
        items, vectors, counters = map(list, zip(*self.__c.dumps[iid]))
        return Item(**{e: f(self, items, vectors, counters) for e, f in extractable.s.items()})

    @extractable
    def items(self, items: List[Item], vectors: np.ndarray, counters: List[Counter]) -> List[Item]:
        return items

    @extractable
    def vectors(self, items: List[Item], vectors: np.ndarray, counters: List[Counter]) -> np.ndarray:
        return vectors

    @extractable
    def counter(self, items: List[Item], vectors: np.ndarray, counters: List[Counter]) -> Counter:
        return reduce(lambda x,y: x+y, counters)

    @extractable
    def center(self, items: List[Item], vectors: np.ndarray, counters: List[Counter]) -> int:
        """
        extract topic

        return Item that center of cluster
        """
        _, index = Extractor.__find_center(vectors)
        return index

    @extractable
    def keywords(self, items: List[Item], vectors: np.ndarray, counters: List[Counter], top: int = 32) -> List[str]:
        """
        extract keywords
        [optional argument top=32 for count of keywords]

        return keywords in cluster
        """
        mean, index = Extractor.__find_center(vectors)
        counter = self.counter(items, vectors, counters)

        def _get_f(t: wID) -> float:
            return float(counter[t])
        def _get_tf(t: wID) -> float:
            max_f = max([_get_f(w) for w in counter.keys() ])
            return .5 + (.5 * _get_f(t) / max_f)
        def _get_idf(t: wID) -> float:
            return 0.01 + normalize(len(self.__c.items)/len(vectors))
        def _get_score(t: wID) -> float:
            return _get_tf(t) * _get_idf(t) + _get_f(t) * 0.001

        words = [(w, _get_score(w)) for w in filter(lambda x: is_noun(x),counter.keys())]
        return sorted(words, key=lambda w: -float(w[1]))[:top]



from KINCluster import settings

import codecs
from itertools import chain
from os import listdir
from os.path import join, dirname, abspath

def __get_stopwords():
    stopwords_dir = join(dirname(__file__), 'stopwords')
    words = set()
    for file_name in listdir(stopwords_dir):
        for word in codecs.open(join(stopwords_dir, file_name), "r", "utf-8").readlines():
            words.add(word)
    return map(lambda x: x.strip(), words)
stopwords = list(__get_stopwords())


from KINCluster.core.item import Item
from KINCluster.lib.tokenizer import tokenizer
from KINCluster import settings

from typing import List, Iterator, Union, Any
from collections import Counter

import numpy as np
from gensim.models import Doc2Vec
from gensim.models.doc2vec import TaggedDocument
from scipy.cluster import hierarchy as hcluster
from scipy.cluster import vq as vq

class Cluster:
    def __init__(self, **kwargs):
        """hyperparameters
            :alpha = learning rate
            :min_alph = minimum learning reate
            :window = max value of vector
            :size = vector size
            :tokenizer = lambda document: str -> list or words: List[str]
        """
        self.settings = kwargs.get('settings', settings)
        alpha = kwargs.get("alpha", self.settings.C_LEARNING_RATE)
        min_alpha = kwargs.get("min_alpha", self.settings.C_LEARNING_RATE_MIN)
        window = kwargs.get("window", self.settings.C_WINDOW)
        size = kwargs.get("size", self.settings.C_SIZE)
        self.trate = kwargs.get("trate", self.settings.C_TRANING_RATE)
        self.epoch = kwargs.get("epoch", self.settings.C_EPOCH)
        self.thresh = kwargs.get("thresh", self.settings.C_THRESHOLD)
        self.tokenizer = tokenizer.s[kwargs.get("tokenizer", settings.C_TOKENIZER)]

        self.model = Doc2Vec(alpha=alpha, min_alpha=min_alpha, window=window, size=size)
        self._items = []
        self._counters = []
        self._vectors = []
        self._clusters = []
        self._dumps = []

    def put_item(self, item: Item):
        self._items.append(item)

    def __vocabs(self) -> Iterator[TaggedDocument]:
        for idx, item in enumerate(self._items):
            token = self.tokenizer(repr(item))
            self._counters.append(Counter(token))
            yield TaggedDocument(token, ['line_%s' % idx])

    def __documents(self) -> Iterator[TaggedDocument]:
        for idx, item in enumerate(self._items):
            yield TaggedDocument(self.tokenizer(str(item)), ['line_%s' % idx])

    def __cluster(self, method, metric, criterion) -> np.ndarray:
        return hcluster.fclusterdata(self._vectors, self.thresh, method=method, metric=metric, criterion=criterion)

    def cluster(self):
        # COMMENT: Top keyword 만 잘라서 분류해보기
        # 
        """cluster process
            : build vocab, using repr of item
            : train items, using str of item
            : get _vectors and _clusters
        """
        self.model.build_vocab(self.__vocabs())

        documents = list(self.__documents())
        for epoch in range(self.epoch):
            self.model.train(documents)
            self.model.alpha *= self.trate
            self.model.min_alpha = self.model.alpha

        self._vectors = np.array(self.model.docvecs)
        self._clusters = self.__cluster(self.settings.C_METHOD, self.settings.C_METRIC, self.settings.C_CRITERION)

        dumps = {c: [] for c in self.unique}
        for c, item, vector, counter in zip(self._clusters, self._items, self._vectors, self._counters):
            dumps[c].append((item, vector, counter))
        self._dumps = list(dumps.values())

    def similar(self, pos, neg=[], top=10):
        return self.model.most_similar(positive=pos,negative=neg,topn=top)

    @property
    def items(self) -> List[Item]:
        return self._items
    @property
    def vocab(self) -> List[str]:
        return self.model.vocab
    @property
    def vocab_count(self) -> List[Counter]:
        return self._counters
    @property
    def dumps(self) -> List[List[Union[Item, np.ndarray]]]:
        return self._dumps
    @property
    def vectors(self) -> np.ndarray:
        return self._vectors
    @property
    def unique(self) -> np.ndarray:
        return np.unique(self._clusters)
    @property
    def clusters(self) -> np.ndarray:
        return self._clusters
    @property
    def distribution(self) -> np.ndarray:
        return Counter(self._clusters)

    def __len__(self):
        return len(self._clusters)


from KINCluster.core.cluster import Cluster 
from KINCluster.core.pipeline import Pipeline 
from KINCluster.core.extractor import Extractor 
from KINCluster.lib.tokenizer import tokenize
from KINCluster import settings

class KINCluster:
    def __init__(self, pipeline, cluster=Cluster, Extractor=Extractor, settings=settings):
        self.pipeline = pipeline
        self.cluster = Cluster(settings=settings)
        self.Extractor = Extractor

    def run(self):
        for item in self.pipeline.capture_item():
            self.cluster.put_item(item)
        self.cluster.cluster()

        extractor = self.Extractor(self.cluster)
        for idx, dump in enumerate(self.cluster.dumps):
            self.pipeline.dress_item(extractor.dump(idx))


# Cluster settings

### Methods = ["linkage", "single", "complete", "average", "weighted", "centroid", "median", "ward"]
C_METHOD="ward"
### Metrics = ["euclidean", "minkowski", "cityblock", "seuclidean", "sqeuclidean", "cosine", "correlation", "hamming", "jaccard", "chebyshev"]
C_METRIC="cosine"
### Criterions = ["inconsistent", "distance", "maxclust", "monocrit", "maxclust_monocrit"]
C_CRITERION="inconsistent"

# alpha
C_LEARNING_RATE = 0.025
# min_alpha
C_LEARNING_RATE_MIN = 0.025
# window
C_WINDOW = 5
# size
C_SIZE = 500
# trate
C_TRANING_RATE = 0.98
# epoch
C_EPOCH = 256
# threshold, recommand between 10-1, 10+1
C_THRESHOLD = 1.15

# tokenizer = ["tokenize", "stemize"]
# or implemeint in lib.tokenizer
C_TOKENIZER = "tokenize"

# Cluster type
## for now, only hcluster
C_TYPE = "hcluster"

# item element setting
ITEM_ELEMENT = [
    'title',
    'content',
    'date',
    'keyword'
]


# Tokenizer setting
TOKEN_POS_TAG = [
]

TOKEN_NEG_TAG = [
]

TOKEN_ZIP_TAG = [
]

# user custom hyperparameter
ITEM_TITLE_RATE     = 10
ITEM_CONTENT_RATE   = 1
ITEM_DATE_RATE      = 5
ITEM_KEYWORD_RATE   = 20



from KINCluster import settings

from collections import OrderedDict

from typing import List, Any

class Item:
    def __init__(self, **kwargs):
        """item elements is ordered dict
        why?: str and repr must return same value always
        """
        self.__e = OrderedDict()
        for k, v in kwargs.items():
            try:
                setattr(self, k, v)
            except:
                raise "do not use item key like 'keys', 'items' or dumplicated key"
            self.__e[k] = v

    def __str__(self) -> str:
        """represent text to clustering
            :apply n-gram or stemize if you need
        """
        return " ".join(map(str, self.values))

    def __repr__(self) -> str:
        """represent vocabs in item
        """
        return " ".join(map(str, self.values))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
           return other.__e == self.__e
        return False 

    @property
    def values(self) -> List[Any]:
        """ return element values
        """
        return self.__e.values()

    @property
    def keys(self) -> List[Any]:
        """ return element keys
        """
        return self.__e.keys()



"""
Cluster core library classes and functions.
"""

__all__ = ['cluster', 'extractor', 'item', 'pipeline']

from KINCluster.core import cluster 
from KINCluster.core import extractor
from KINCluster.core import item
from KINCluster.core import pipeline


"""
KINCluster is clustering like KIN.

release note: 
- version 0.1.5.3
    now custom setting available.
    see settings.py

- version 0.1.5.2
    change item, extractor, pipeline module
    now, pipeline.dress_item pass just item(extractor.dump)
    fix prev versions error (too many value to unpack)

- version 0.1.5.1
    rebase module, just import KINCluster
    simplify extractor module

- version 0.1.4.5
    First release
"""

__version__ = '0.1.5.3'

__all__ = ['KINCluster', 
'Cluster', 'Extractor', 'Item', 'Pipeline', 
'tokenizer', 'stopwords']

from KINCluster.KINCluster import KINCluster

from KINCluster.core.cluster import Cluster 
from KINCluster.core.extractor import Extractor 
from KINCluster.core.item import Item 
from KINCluster.core.pipeline import Pipeline 

from KINCluster.lib.tokenizer import tokenizer
from KINCluster.lib.stopwords import stopwords


from KINCluster.core.item import Item
from KINCluster.lib import tokenizer
from KINCluster import settings

from typing import List, Iterator, Union, Any
nparray = Any

class Pipeline:
    def __init__(self):
        pass

    def capture_item(self) -> Iterator[Item]:
        """grab items
            :must be item generater
        """
        raise Exception('Override capture_item function to generator<Item>')
        pass

    def dress_item(self, item: Item):
        """dress item
            :item (has element extractable)
                - items     : cluster items
                - vectors   : cluster vectors
                - counter   : cluster counter
        """
        pass

    def __finalize__(self):
        pass


from KINCluster.lib.stopwords import stopwords
from KINCluster import settings

from typing import List, Union
from itertools import chain
import re

# import os
# if os.name == 'nt':
from konlpy.tag import Twitter
tagger = Twitter()
# else:
#     from konlpy.tag import Mecab
#     tagger = Mecab()

# type hinting
TAG = str

class tokenizer:
    s = {}
    def __init__(self, func):
        tokenizer.s[func.__name__] = func
        self.func = func
    def __get__(self, obj, klass=None):
        def _call_(*args, **kwargs):
            return self.func(obj, *args, **kwargs)
        return _call_

pos_tag = settings.TOKEN_POS_TAG
neg_tag = settings.TOKEN_NEG_TAG
zip_tag = settings.TOKEN_ZIP_TAG
zip_token = "**//*///**/*//*//*"
def tokenizer_init():
    # pos_tag = []
    # neg_tag = []
    # zip_tag = [[]]
    # zip_token = ""
    pass

def filter_tag(text, pos_tag : List[TAG] = pos_tag, neg_tag : List[TAG] = neg_tag) -> str:
    # negative filter
    if not pos_tag:
        return " ".join([w for w, t in tagging(text) if not t in neg_tag])
    # positive fileter
    else:
        return " ".join([w for w, t in tagging(text) if t in pos_tag])

# find quotation for 'important word'
pat_small_quot = re.compile(u"\'(.+?)\'")
pat_double_quot = re.compile(u"\"(.+?)\"")
def find_quotations(text):
    mat_small = pat_small_quot.finditer(text)
    mat_double = pat_double_quot.finditer(text)
    return list(mat_small) + list(mat_double)

def is_noun(word):
    w, t = tagging(word)[0]
    return t[0] == 'N'

@tokenizer
def tokenize(text) -> List[str]:
    return [word for word in text.split() if not word in stopwords]

# konlpy custom dic 써보기 
# 
@tokenizer
def stemize(text, pos_tag : List[TAG] = pos_tag, neg_tag : List[TAG] = neg_tag) -> List[str]:
    def extract_quotations(text) -> Union[List[str], str]:
        matches = []
        c = 0
        for match in find_quotations(text):
            text = " ".join([text[:match.start() + c], zip_token, text[match.end() + c:]])
            c += len(zip_token) - len(match.group()) + 2
            matches.append(match.group()[1:-1])
        return (matches, text)

    def zip_tokens(tokens):
        for zip_pat in zip_tag:
            if tokens[0][1] in zip_pat and tokens[-1][1] in zip_pat:
                words, tokens = map(list, zip(*tokens))
                tokens = [("".join(words), 'ZIP')]
        for word in list(zip(*tokens))[0]:
            if not word in stopwords:
                yield word

    # return [tokens[0] for tokens in tagger.pos(text) if tokens[1][0] == 'N' ]
    matches, text = extract_quotations(text)
    words = [zip_tokens(tokens) for tokens in [tagger.pos(word) for word in text.split()]]
    ret = list(chain.from_iterable(words))
    return [r == zip_token and matches.pop() or r for r in ret]

def tagging(text) -> List[Union[str, TAG]]:
    return tagger.pos(text)

tokenizer_init()


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from KINCluster import __version__

setup(
    name = 'KINCluster',
    packages = ['KINCluster' ,'KINCluster/core', 'KINCluster/lib'],
    include_package_data=True,
    version = __version__,
    description = 'Korean Involute News Cluster',
    license = 'MIT',
    author = 'Bae jiun',
    author_email = 'alice.maydev@gmail.com',
    
    url = 'https://github.com/MaybeS/KINCluster',
    keywords = ['KINCluster', 'cluster', 'documents' 'doc2vec', 'tokenize', 'korean'],

    install_requires=[
        'gensim==0.13.4.1',
        'konlpy>=0.4.4',
        'scipy>=0.19.0',
        'numpy>=1.12.1+mkl'
    ],

    classifiers=(
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ),

    entry_points={
        'console_scripts': [
            'jikji = jikji.cli:main',
        ],
    },
)